/**
 * @file
 * Contains the javascript implementation of the dynamic entitylist widget.
 *
 * @package Drupal\blizz_dynamic_entitylist
 */

(function ($, Drupal, drupalSettings) {

  Drupal.BlizzDynamicEntitylistWidget = {

    getAllAttributes: function (elem, filter) {

      // "filter" is an optional parameter.
      if (typeof filter === 'undefined') {
        filter = false;
      }

      // Prepare an object holding the results.
      var attributes = [];

      // Determine all attributes and their respective values.
      $(elem).each(function () {
        $.each(this.attributes, function () {
          if (this.specified) {
            attributes[this.name] = this.value;
          }
        });
      });

      // If there is a filter set, use it.
      if (filter !== false) {
        var unfiltered = $.extend([], attributes);
        attributes = [];
        for (var key in unfiltered) {
          if (unfiltered.hasOwnProperty(key) && key.match(new RegExp(filter))) {
            attributes[key] = unfiltered[key];
          }
        }
      }

      // Return the result.
      return attributes;
    },

    bindEvents: function (elem) {
      var classes = $(elem).attr('class').split(' '), callback = false, arg;
      for (var index in classes) {
        switch (true) {
          case (/^select-/).test(classes[index]):
            callback = 'ajaxSelectOnChange';
            arg = classes[index].split('-').pop();
            break;

          case (/^autocomplete-/).test(classes[index]):
            callback = 'autocomplete';
            arg = classes[index].split('-').pop();
            break;

          case (/^data-selector/).test(classes[index]):
            callback = 'selectOnChange';
            arg = null;
            break;

          case (/^data-input/).test(classes[index]):
            callback = 'onChange';
            arg = null;
            break;

        }
        if (callback) {
          break;
        }
      }

      if (callback !== false) {
        Drupal.BlizzDynamicEntitylistWidget[callback](elem, arg);
      }
    },

    ajaxSelectOnChange: function (elem, argument) {
      $(elem).bind(
        'change',
        function () {

          // Determine the DOM context of the select changed.
          var $fieldContainer = $(this).parents().filter('.field--type-blizz-dynamic-entitylist');

          // Which element should hold the results?
          var targetarea = $(this).attr('data-targetarea');

          if ($(this).val() !== '') {

            // Determine the URL to make a call to.
            var url = drupalSettings.blizz_dynamic_entitylist['callback-urls'][argument];

            // Are there any placeholders needing to be replaced with actual values?
            var placeholders = Drupal.BlizzDynamicEntitylistWidget.getAllAttributes(elem, '^data\-replace\-');

            // Process each placeholder defined...
            for (var index in placeholders) {
              if (placeholders.hasOwnProperty(index)) {
                // Determine the actual string of the placeholder.
                var placeholder = index.replace('data-replace-', '').toUpperCase();

                // Replace the placeholder string in the URL with the actual value.
                url = url.replace(placeholder, $(placeholders[index], $fieldContainer).val())
              }
            }

            // Make the ajax call.
            $.ajax({
              url: url,
              dataType: 'html',
              success: function (data) {
                // Insert the response into the DOM.
                $('.' + targetarea, $fieldContainer).html(data);
                // Re-initialise any elements that need ajax functionalities.
                Drupal.BlizzDynamicEntitylistWidget.init();
              },
              complete: function () {
                // If there is an element contained in the response stating to auto-trigger
                // the next step, then auto-trigger and remove the attribute.
                $('*[data-trigger-next-step="1"]').each(function (index, elem) {
                  $(elem).trigger('change').removeAttr('data-trigger-next-step');
                });
              }
            });
          }
          else {
            // Empty the target container if the element's value is switched back to empty.
            $('.' + targetarea, $fieldContainer).html('');
          }

          // Refresh list definition after selection change.
          Drupal.BlizzDynamicEntitylistWidget.serializeWidget(
            $(this).parents().filter('.field--type-blizz-dynamic-entitylist')
          );

        }
      );
    },

    selectOnChange: function (elem) {
      // Refresh list definition after selection change.
      $(elem).bind(
        'change',
        function () {
          Drupal.BlizzDynamicEntitylistWidget.serializeWidget(
            $(this).parents().filter('.field--type-blizz-dynamic-entitylist')
          );
        }
      );
    },

    autocomplete: function (elem, argument) {

      // Determine the URL to make a call to.
      var url = drupalSettings.blizz_dynamic_entitylist['callback-urls']['autocomplete'];
      url = url.replace('ENTITY_TYPE', argument);

      var multipleSelect = (Number($(elem).attr('data-multiple')) === 1);
      var options = {
            width: '300px',
            multiple: multipleSelect,
            allowClear: true,
            placeholder: '',
            position: 'below',
            ajax: {
              url: url.replace('BUNDLE', $(elem).attr('data-targetbundles')),
              dataType: 'json',
              data: function (params) {
                return {
                    term: params.term
                };
              }
            }
          };

      if (multipleSelect) {
        options.maximumSelectionLength = Number($(elem).attr('data-maximumselectionlength'));
      }

      // Bind the autocomplete event / select2 library to the select field.
      $(elem).select2(options);

      // Refresh list definition after selection change.
      $(elem).bind(
        'change',
        function () {
          Drupal.BlizzDynamicEntitylistWidget.serializeWidget(
            $(this).parents().filter('.field--type-blizz-dynamic-entitylist')
          );
        }
      );

    },

    onChange: function (elem) {
      // Refresh list definition after value change.
      $(elem).bind(
        'change',
        function () {
          Drupal.BlizzDynamicEntitylistWidget.serializeWidget(
            $(this).parents().filter('.field--type-blizz-dynamic-entitylist')
          );
        }
      );
    },

    serializeWidget: function (elem) {
      var $listdefinition = $('.blizzdynamicentitylist-definition', elem);
      var definition = {
        entity_type_id: $('select.select-bundlelist', elem).val(),
        bundle: $('select.select-viewmodelist', elem).val(),
        viewmode: $('select.list-viewmode', elem).val(),
        length: $('input.list-length', elem).val(),
        offset: $('input.list-offset', elem).val(),
        filters: {}
      };
      $('.filters select.filtervalue', elem).each(function (index, elem) {
        definition.filters[$(elem).attr('data-sourcefield')] = $(elem).val();
      });
      $listdefinition.val(JSON.stringify(definition));

      if (definition.entity_type_id && definition.bundle) {
        $.ajax({
          url: drupalSettings.blizz_dynamic_entitylist['callback-urls'].preview,
          dataType: 'html',
          cache: false,
          method: 'post',
          data: definition,
          success: function (preview) {
            $('.preview', $(elem)).html(preview);
          }
        });
      }
    },

    init: function () {

      /*
        Since this widget relies upon an external library, we will make sure this
        library is loaded before initializing the widget's fields.

        Reason:
        It may be possible, that this field is initially rendered in some kind of
        a "preview" mode when editing it's host entity, so this field widget library
        will get loaded by a subsequent ajax request when entering the "edit" mode
        for this field. If so, it seems that external libraries like the "Select2"
        library defined in this module will not get loaded properly and therefore
        this widget will throw a javascript error, resulting in useless select fields.
       */
      if (typeof $.fn.select2 === 'undefined') {
        $.ajax({
          async: false,
          url: 'https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.6-rc.0/js/select2.min.js',
          dataType: 'script',
          success: function () {
            Drupal.BlizzDynamicEntitylistWidget.init();
          }
        });
        $('<link>')
          .appendTo('head')
          .attr({
              type: 'text/css',
              rel: 'stylesheet',
              href: 'https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.6-rc.0/css/select2.min.css'
          });
      }
      else {
        $('.field--type-blizz-dynamic-entitylist .blizzdynamicentitylist-selector')
          .once('bindEvents')
          .each(function (index, elem) {

            // Bind input functionality.
            Drupal.BlizzDynamicEntitylistWidget.bindEvents(elem);

            // Prepopulate further widget steps, if there is only a single entity type
            // available for selection and no default field values are present.
            $('*[data-trigger-next-step="1"]').each(function (index, elem) {
              var $bundleSelect = $('select.select-viewmodelist', $(elem).parents().filter('.field--type-blizz-dynamic-entitylist'));
              if ($bundleSelect.length < 1 || ($bundleSelect.length && $bundleSelect.val() === '')) {
                $(elem).trigger('change').removeAttr('data-trigger-next-step');
              }
            });

          });
      }

    }

  };

  Drupal.behaviors.BlizzDynamicEntitylist = {
    attach: function () {
      Drupal.BlizzDynamicEntitylistWidget.init();
    }
  };

}(jQuery, Drupal, drupalSettings));
