<?php

namespace Drupal\blizz_dynamic_entitylist\Plugin\Field\FieldType;

use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Plugin implementation of the 'blizz_dynamic_entitylist' field type.
 *
 * @FieldType(
 *   id = "blizz_dynamic_entitylist",
 *   label = @Translation("Dynamic Entity list"),
 *   description = @Translation("Provides a field type that lets you define automatic lists of entities based upon selection criteria."),
 *   default_widget = "blizz_dynamic_entitylist_widget",
 *   default_formatter = "blizz_dynamic_entitylist_formatter"
 * )
 */
class BlizzDynamicEntitylist extends FieldItemBase {

  /**
   * Custom service to easy the handling of entities.
   *
   * @var \Drupal\blizz_dynamic_entitylist\Services\EntityServicesInterface
   */
  protected $entityServices;

  /**
   * {@inheritdoc}
   */
  public function __construct(DataDefinitionInterface $definition, $name = NULL, TypedDataInterface $parent = NULL) {
    parent::__construct($definition, $name, $parent);
    $this->entityServices = \Drupal::service('blizz_dynamic_entitylist.entity_services');
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultStorageSettings() {
    return array(
      'entity_types' => [],
    ) + parent::defaultStorageSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function storageSettingsForm(array &$form, FormStateInterface $form_state, $has_data) {
    $settings = $this->getSettings();
    return [
      'entity_types' => [
        '#type' => 'checkboxes',
        '#title' => $this->t('Limit lists to these entity types.'),
        '#description' => $this->t('Leave unchecked to make all content entity types listable.'),
        '#options' => $this->entityServices->getContentEntityTypeOptions(),
        '#default_value' => !empty($settings['entity_types']) ? $settings['entity_types'] : [],
        '#disabled' => $has_data,
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function storageSettingsToConfigData(array $settings) {
    $settings['entity_types'] = array_values(array_filter($settings['entity_types']));
    return parent::storageSettingsToConfigData($settings);
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $properties['listdefinition'] = DataDefinition::create('string')
      ->setLabel(new TranslatableMarkup('The (serialized) list definition of this entity list instance.'))
      ->setRequired(TRUE);

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    return [
      'columns' => [
        'listdefinition' => [
          'description' => 'The (serialized) list definition of this entity list instance.',
          'type' => 'blob',
          'length' => 'normal',
          'serialize' => TRUE,
          'not null' => TRUE,
          'binary' => TRUE,
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty() {
    return empty($this->get('listdefinition')->getValue());
  }

}
