<?php

namespace Drupal\blizz_dynamic_entitylist\Services;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Symfony\Component\HttpFoundation\Response;
use Zend\Diactoros\Response\JsonResponse;

/**
 * Class BlizzDynamicEntitylistWidgetServices.
 *
 * @package Drupal\blizz_dynamic_entitylist\Services
 */
class BlizzDynamicEntitylistWidgetServices implements BlizzDynamicEntitylistWidgetServicesInterface {

  /**
   * Drupal's rendering service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * Drupal's string translation service.
   *
   * @var \Drupal\Core\StringTranslation\TranslationInterface
   */
  protected $stringTranslator;

  /**
   * Custom service to easy the handling of entities.
   *
   * @var \Drupal\blizz_dynamic_entitylist\Services\EntityServicesInterface
   */
  protected $entityServices;

  /**
   * BlizzDynamicEntitylistWidgetServices constructor.
   *
   * Provides functionality for different steps of the field widget.
   *
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   Drupal's rendering service.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translator
   *   Drupal's string translation service.
   * @param \Drupal\blizz_dynamic_entitylist\Services\EntityServicesInterface $entity_services
   *   Custom service to easy the handling of entities.
   */
  public function __construct(
    RendererInterface $renderer,
    TranslationInterface $string_translator,
    EntityServicesInterface $entity_services
  ) {
    $this->renderer = $renderer;
    $this->stringTranslator = $string_translator;
    $this->entityServices = $entity_services;
  }

  /**
   * {@inheritdoc}
   */
  public function widgetStage2($entity_type, array $listdefinition = [], $raw = FALSE) {

    // Get an array holding all bundles of the biven entity type id.
    $bundles = $this->entityServices->getEntityTypeBundleOptions($entity_type);
    if (count($bundles) > 1) {
      $bundles = [
        '' => $this->stringTranslator->translate('Please choose'),
      ] + $bundles;
    }

    $elem = [
      '#type' => 'container',
      'bundleselect' => [
        '#type' => 'select',
        '#title' => $this->stringTranslator->translate('Bundle'),
        '#description' => $this->stringTranslator->translate('Please select the bundle you would like to get a list of.'),
        '#description_display' => 'after',
        '#options' => $bundles,
        '#default_value' => isset($listdefinition['bundle']) ? $listdefinition['bundle'] : NULL,
        '#required' => TRUE,
        '#attributes' => [
          'class' => [
            'blizzdynamicentitylist-selector',
            'select-viewmodelist',
          ],
          'data-targetarea' => 'viewmode',
          'data-replace-ENTITY_TYPE' => 'select.select-bundlelist',
          'data-replace-BUNDLE' => 'select.select-viewmodelist',
        ],
      ],
    ];

    if (empty($listdefinition)) {
      $elem['viewmode'] = [
        '#type' => 'markup',
        '#markup' => '<div class="viewmode"></div>',
      ];
    }
    else {
      $elem['viewmode'] = $this->widgetStage3(
        $listdefinition['entity_type_id'],
        $listdefinition['bundle'],
        $listdefinition,
        TRUE
      ) + ['#prefix' => '<div class="viewmode">', '#suffix' => '</div>'];
    }

    if (count($bundles) == 1 && !$raw) {
      $elem['bundleselect']['#attributes']['data-trigger-next-step'] = 1;
    }

    return $raw ? $elem : new Response($this->renderer->render($elem));
  }

  /**
   * {@inheritdoc}
   */
  public function widgetStage3($entity_type, $bundle, array $listdefinition = [], $raw = FALSE) {

    // Get an array of all available view modes for a given
    // entity type/bundle combination.
    $viewmodes = $this->entityServices->getEntityTypeBundleViewmodes($entity_type, $bundle);
    if (count($viewmodes) > 1) {
      $viewmodes = [
        '' => $this->stringTranslator->translate('Please choose'),
      ] + $viewmodes;
    }
    elseif (empty($viewmodes)) {
      $viewmodes = [
        'default' => $this->stringTranslator->translate('Default'),
      ];
    }

    // Build the form potion of the third widget step.
    $elem = [
      '#type' => 'container',
      // View mode selection.
      'viewmodeselect' => [
        '#type' => 'select',
        '#title' => $this->stringTranslator->translate('View mode'),
        '#description' => $this->stringTranslator->translate('Please select the view mode you would like your list to be rendered in.'),
        '#description_display' => 'after',
        '#options' => $viewmodes,
        '#default_value' => isset($listdefinition['viewmode']) ? $listdefinition['viewmode'] : NULL,
        '#required' => TRUE,
        '#attributes' => [
          'class' => [
            'blizzdynamicentitylist-selector',
            'data-selector',
            'list-viewmode',
          ],
        ],
      ],
      // List length.
      'length' => [
        '#type' => 'number',
        '#title' => $this->stringTranslator->translate('Length of the list'),
        '#description' => $this->stringTranslator->translate('How many entities would you like to include in your list? (0 = infinite)'),
        '#description_display' => 'after',
        '#required' => TRUE,
        '#min' => 0,
        '#value' => isset($listdefinition['length']) ? $listdefinition['length'] : 10,
        '#attributes' => [
          'class' => [
            'blizzdynamicentitylist-selector',
            'data-input',
            'list-length',
          ],
        ],
      ],
      // List offset.
      'offset' => [
        '#type' => 'number',
        '#title' => $this->stringTranslator->translate('Offset'),
        '#description' => $this->stringTranslator->translate('How many entities would you like to skip at the beginning of your list? (0 = none)'),
        '#description_display' => 'after',
        '#required' => TRUE,
        '#min' => 0,
        '#value' => isset($listdefinition['offset']) ? $listdefinition['offset'] : 0,
        '#attributes' => [
          'class' => [
            'blizzdynamicentitylist-selector',
            'data-input',
            'list-offset',
          ],
        ],
      ],
      'filters' => ['#access' => FALSE],
      'preview' => [
        '#type' => 'details',
        '#title' => $this->stringTranslator->translate('Preview'),
        '#open' => FALSE,
        'previewarea' => [
          '#type' => 'markup',
          '#markup' => '<div class="preview"></div>',
        ],
      ],
    ];

    // Taxonomy term reference fields will be defined like this:
    $taxonomyReferenceFieldCriteria = [
      'getType' => 'entity_reference',
      'getSetting' => [
        'arguments' => ['handler'],
        'value' => 'default:taxonomy_term',
      ],
    ];

    // Are there any taxonomy term reference fields within the selected
    // entity type/bundle combination?
    $taxonomyReferenceFields = $this->entityServices->getEntityTypeBundleFields(
      $entity_type,
      $bundle,
      $taxonomyReferenceFieldCriteria
    );

    // If there are taxRef fields, generate filter fields as well.
    if (!empty($taxonomyReferenceFields)) {

      // Fetch the labels of the fields.
      $taxonomyReferenceFieldOptions = $this->entityServices->getEntityTypeBundleFieldOptions(
        $entity_type,
        $bundle,
        $taxonomyReferenceFieldCriteria
      );

      // Set a container holding the different filter fields.
      $elem['filters'] = [
        '#type' => 'details',
        '#open' => FALSE,
        '#title' => $this->stringTranslator->translate('Filters'),
        '#description' => $this->stringTranslator->translate('Should there be any filter criteria applied to the entities selected?'),
        'filtercontainer' => [
          '#type' => 'container',
          '#attributes' => [
            'class' => [
              'filters',
            ],
          ],
        ],
      ];

      // Generate the filter fields.
      foreach ($taxonomyReferenceFieldOptions as $field_machine_name => $label) {

        // Get the field definition.
        $field_definition = $taxonomyReferenceFields[$field_machine_name];

        // Determine the cardinality of the field.
        $cardinality = $field_definition->getFieldStorageDefinition()->getCardinality();

        // Can this field hold more than a single value?
        $multiple = ($cardinality == -1 || $cardinality > 1) ? 1 : 0;

        // Which target bundles are targeted?
        $target_bundles = $field_definition->getSetting('handler_settings')['target_bundles'];

        // Add the filter field definition.
        $elem['filters']['filtercontainer'][$field_machine_name] = [
          '#type' => 'container',
          '#attributes' => [
            'class' => [
              'filtercontainer',
            ],
          ],
          'filtervalue' => [
            '#type' => 'select',
            // Intentionally no field name, explanation see below.
            '#name' => FALSE,
            '#title' => $label,
            '#attributes' => [
              'class' => [
                'blizzdynamicentitylist-selector',
                'autocomplete-taxonomy_term',
                'filtervalue',
              ],
              'data-sourcefield' => $field_machine_name,
              'data-targetbundles' => implode(', ', $target_bundles),
              'data-multiple' => $multiple,
              'data-maximumselectionlength' => $cardinality,
            ],
          ],
        ];

        /*
         * So why do the 'filtervalue' fields use no field name?
         *
         * Simply: they are not readily built at the time of the form
         * generation, instead options get filled in via custom Ajax calls.
         *
         * So, if an option not present at the time of the form generation would
         * be chosen, Drupal would detect an error ("An illegal choice has been
         * detected. Please contact the site administrator.") at the time of
         * form submission.
         *
         * By denying this field a field name, this field will simply not get
         * transmitted within the request headers and Drupal will not try to
         * validate it. Since the widget gets serialized and this field's single
         * value is not of importance, we can just ignore it.
         */

        // If the field cardinality allows for multiple values, ...
        if ($multiple) {

          // ...set an appropriate html attribute.
          $elem['filters']['filtercontainer'][$field_machine_name]['filtervalue']['#attributes']['multiple'] = 'multiple';

        }

        // Prepopulate filtervalue fields with potential data.
        if (!empty($listdefinition['filters'][$field_machine_name])) {

          // Shortcut for convenience...
          $input_element = &$elem['filters']['filtercontainer'][$field_machine_name]['filtervalue'];

          // Load the referenced entities to generate complete field options.
          $terms = $this->entityServices->loadEntities(
            'taxonomy_term',
            is_array($listdefinition['filters'][$field_machine_name])
              ? $listdefinition['filters'][$field_machine_name]
              : [$listdefinition['filters'][$field_machine_name]]
          );

          // Generate the chosen field options.
          $input_element['#options'] = array_map(
            function (ContentEntityInterface $term) {
              return $term->label();
            },
            $terms
          );

          // Set the chosen options also as default values.
          $input_element['#default_value'] = is_array($listdefinition['filters'][$field_machine_name])
            ? array_combine(
                array_values($listdefinition['filters'][$field_machine_name]),
                array_values($listdefinition['filters'][$field_machine_name])
              )
            : $listdefinition['filters'][$field_machine_name];

        }
      }

    }

    return $raw ? $elem : new Response($this->renderer->render($elem));
  }

  /**
   * {@inheritdoc}
   */
  public function handleAutocomplete($entity_type, $bundle) {
    $entityBundleOptions = $this->entityServices->getEntityTypeBundleOptions($entity_type);
    $query = $this->entityServices->getQuery($entity_type);
    if ($bundlekey = $this->entityServices->getEntityKey($entity_type, 'bundle')) {
      $query->condition($bundlekey, explode(',', $bundle), 'IN');
      $query->sort($bundlekey, 'ASC');
    }
    if ($labelkey = $this->entityServices->getEntityKey($entity_type, 'label')) {
      if (!empty($_GET['term'])) {
        $query->condition($labelkey, sprintf('%%%s%%', $_GET['term']), 'LIKE');
      }
      $query->sort($labelkey, 'ASC');
    }
    $ids = $query->execute();
    $entities = $this->entityServices->loadEntities($entity_type, $ids);

    $bundles = array_unique(array_map(
      function (ContentEntityInterface $entity) {
        return $entity->bundle();
      },
      $entities
    ));

    $results = [];
    foreach ($bundles as $bundle) {
      $optgroup = [
        'text' => $entityBundleOptions[$bundle],
        'children' => array_values(array_map(
          function (ContentEntityInterface $entity) {
            return [
              'id' => $entity->id(),
              'text' => $entity->label(),
            ];
          },
          array_filter(
            $entities,
            function (ContentEntityInterface $entity) use ($bundle) {
              return $entity->bundle() == $bundle;
            }
          )
        )),
      ];
      $results[] = $optgroup;
    }

    return new JsonResponse([
      'results' => array_values($results) ?: [],
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getListDefinition(array $fieldcontents) {
    if (!empty($fieldcontents) && is_array($fieldcontents) && isset($fieldcontents['listdefinition'])) {
      if ($listdefinition = unserialize($fieldcontents['listdefinition'])) {
        $listdefinition = array_shift($listdefinition);
        if (empty(array_diff_key(
          array_flip([
            'entity_type_id',
            'bundle',
            'viewmode',
            'length',
            'offset',
            'filters',
          ]),
          $listdefinition))
        ) {
          return $listdefinition;
        }
      }
    }
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getListEntities(array $listdefinition, $preview_mode = FALSE) {

    // Fetch the antity definition.
    $entity_bundle_fields = $this->entityServices->getEntityTypeBundleFields(
      $listdefinition['entity_type_id'],
      $listdefinition['bundle'],
      [],
      FALSE
    );

    // Create the entity query.
    $query = $this->entityServices->getQuery($listdefinition['entity_type_id']);

    // Apply the bundle condition.
    if ($bundlekey = $this->entityServices->getEntityKey($listdefinition['entity_type_id'], 'bundle')) {
      $query->condition(
        $bundlekey,
        $listdefinition['bundle'],
        '='
      );
    }

    // If the entities feature a "published" flag, load just published entities.
    if ($statuskey = $this->entityServices->getEntityKey($listdefinition['entity_type_id'], 'status')) {
      $query->condition(
        $statuskey,
        1,
        '='
      );
    }

    // Determine the maximum number of entities requested.
    if ($preview_mode) {
      // Max. 10 entities in preview mode...
      $length = ($listdefinition['length'] != 0 && $listdefinition['length'] < 10) ? $listdefinition['length'] : 10;
    }
    else {
      $length = $listdefinition['length'] != 0 ? $listdefinition['length'] : 999999;
    }

    // Apply offset and length conditions.
    $query->range(
      $listdefinition['offset'],
      $length
    );

    // Apply any filters defined, but only if the fields actually exist.
    if (!empty($listdefinition['filters'])) {
      foreach ($listdefinition['filters'] as $field => $value) {
        if (isset($entity_bundle_fields[$field]) && !empty($value)) {
          $query->condition(
            $field,
            $value,
            is_array($value) ? 'IN' : '='
          );
        }
      }
    }

    // Sorting.
    if (isset($entity_bundle_fields['created'])) {
      // Order the list to show newest entities first.
      $query->sort('created', 'DESC');
    }
    else {
      // If the entity type does not feature a "created" timestamp, we'll
      // sort the list based upon the entity label.
      $query->sort(
        $this->entityServices->getEntityKey($listdefinition['entity_type_id'], 'label'),
        'ASC'
      );
    }

    // Perform an access check on matching entities.
    $query->accessCheck(TRUE);

    // Execute the query and return the results.
    return $query->execute();

  }

  /**
   * {@inheritdoc}
   */
  public function getListPreviewByAjax() {
    $listdefinition = [
      'entity_type_id' => $_POST['entity_type_id'],
      'bundle' => $_POST['bundle'],
      'length' => isset($_POST['length']) ? (int) $_POST['length'] : 10,
      'offset' => isset($_POST['offset']) ? (int) $_POST['offset'] : 0,
      'filters' => [],
    ];

    foreach ($_POST['filters'] as $field => $value) {
      $listdefinition['filters'][$field] = $value;
    }

    $ids = $this->getListEntities($listdefinition, TRUE);
    $entities = $this->entityServices->loadEntities($listdefinition['entity_type_id'], $ids);

    return $this->listPreview($_POST['entity_type_id'], $_POST['bundle'], $entities);
  }

  /**
   * {@inheritdoc}
   */
  public function listPreview($entity_type_id, $bundle, array $entities = [], $raw = FALSE) {
    array_walk(
      $entities,
      function (ContentEntityInterface &$value, $key) {
        $entity = $value;
        $value = [
          $entity->toLink($entity->label(), 'canonical', ['absolute' => TRUE, 'attributes' => ['target' => '_blank']]),
          $key,
        ];
      }
    );

    $entity_type_options = $this->entityServices->getContentEntityTypeOptions();
    $bundleOptions = $this->entityServices->getEntityTypeBundleOptions($entity_type_id);

    $table = [
      '#type' => 'table',
      '#header' => [
        $this->stringTranslator->translate('Entity label'),
        $this->stringTranslator->translate('Entity ID'),
      ],
      '#rows' => array_values($entities),
      '#cache' => [
        'max-age' => 0,
      ],
      '#prefix' => sprintf(
        '<div>%s</div>',
        $this->stringTranslator->translate(
          'Listing Entities of type @entity_type:@bundle',
          [
            '@entity_type' => $entity_type_options[$entity_type_id],
            '@bundle' => $bundleOptions[$bundle],
          ]
        )
      ),
      '#suffix' => sprintf(
        '<div>%s</div>',
        $this->stringTranslator->translate('Preview mode: displaying max. 10 items')
      ),
    ];

    return $raw ? $table : new Response($this->renderer->render($table));
  }

}
